from pathlib import Path


FILENAME_CANDIDATES = ['.zshrc', '.bashrc']
FILEPATH_CANDIDATES = [Path.home() / Path(fn) for fn in FILENAME_CANDIDATES]

TODO_AUTOCOMPLETION_MARK = (
	"# todocli autocompletion (this was added by: todo --install-autocompletion)"
)


def install_autocompletion():
	config_filepath = None
	for filepath in FILEPATH_CANDIDATES:
		if filepath.exists():
			config_filepath = filepath
			break

	if config_filepath is None:
		print(
			"No appropriate config file was found. todo looks for the "
			"following files: "
			"{}".format(', '.join(str(p) for p in FILEPATH_CANDIDATES))
		)
		return

	already_there = False
	with open(str(config_filepath)) as config_file:
		for line in config_file:
			if line.strip() == TODO_AUTOCOMPLETION_MARK:
				already_there = True
				break

	if already_there:
		print(
			"Autocompletion is already installed in config file "
			"{}".format(str(config_filepath))
		)
		return

	autocompletion_payload = get_autocompletion_payload()

	with open(str(config_filepath), 'a') as config_file:
		config_file.write('\n\n')
		config_file.write(TODO_AUTOCOMPLETION_MARK + '\n')
		config_file.write(autocompletion_payload)

	print(
		"Autocompletion installed in config file {}\n"
		"You may `source {}` to enjoy it on this terminal.".format(
			config_filepath, config_filepath,
		)
	)


def get_autocompletion_payload():
	import pkg_resources

	filename = pkg_resources.resource_filename('todo.bash_completion', 'toduh.sh')
	with open(filename) as f:
		payload = f.read()
	return payload



import sqlite3, json, os
import os.path as op
from datetime import datetime

from . import utils, init_db
from .utils import DATA_DIR, DB_PATH, DATAFILE_NAME, DATA_CTX_NAME

DATETIME_MIN = '0001-01-01 00:00:00'
END_OF_JSON = '2.1'


def setup_data_access(current_version):
	"""
	Prepare the sqlite database so that it's ready to be used by the
	application. It's supposed to work in any environment (new installation,
	old version, etc) and will make any necessary conversion between different
	versions (for example converting the json datafile from v2.2- into a
	sqlite database)
	"""
	if not op.exists(DATA_DIR):
		os.makedirs(DATA_DIR)

	init_db.update_database(DB_PATH, current_version)
	if current_version is not None \
	and utils.compare_versions(current_version, END_OF_JSON) <= 0:
		json_path = op.join(DATA_DIR, DATAFILE_NAME)
		with open(json_path) as datafile:
			data = json.load(datafile)
		connection = sqlite3.connect(DB_PATH)
		transfer_data(connection, data)


def transfer_data(connection, data):
	""" Transfer all data from a v2.2- JSON datafile held in the `data`
	dictionary into a sqlite database connected to with `connection`."""
	daccess = DataAccess(connection)
	for ctx, props in data['contexts'].items():
		ctx = dbfy_context(ctx)
		options = []
		if 'p' in props:
			options.append(('priority', props['p']))
		if 'v' in props and props['v'] == 'hidden':
			options.append(('visibility', 'hidden'))
		daccess.get_or_create_context(ctx, options)
	for task in data['tasks']:
		title = task['content']
		if 'context' in task:
			ctx = dbfy_context(task['context'])
			daccess.get_or_create_context(ctx)
		else:
			ctx = ''

		options = []
		if 'created' in task:
			created = iso2sqlite(task['created'])
		else:
			created = DATETIME_MIN
		options.append(('created', created))
		for prop in ['start', 'deadline']:
			if prop in task:
				options.append((prop, iso2sqlite(task[prop])))
		if 'done' in task and task['done']:
			options.append(('done', '1'))
		if 'priority' in task:
			options.append(('priority', task['priority']))
		daccess.add_task(title, None, ctx, options)
	daccess.exit()


# In the database, contexts all descent from the root context ('') and any
# non-root context therefore starts with a dot as the dot separates
# different hierarchical levels (<empty string> <dot> <subcontext name>).
# For convenience, we allow the user to type a context's path without the
# starting dot. To do this, we manually add a dot at the begining of path
# that doesn't already start with a dot, unless the path is empty in which
# case it explicitly designates the root context.
#
# dbfy_context and userify_context perform necessary convertions between user-
# typed contexts' path and their DB representation.

def dbfy_context(ctx):
	if ctx == '' or ctx.startswith('.'):
		return ctx
	else:
		return '.'+ctx


def userify_context(ctx):
	if ctx == '':
		return ''
	else:
		return ctx[1:]

# This one is used by transfer_data
def iso2sqlite(iso_date):
	dt = datetime.strptime(iso_date, utils.ISO_DATE)
	return dt.strftime(utils.SQLITE_DT_FORMAT)


def get_insert_components(options):
	""" Takes a list of 2-tuple in the form (option, value) and returns a
	triplet (colnames, placeholders, values) that permits making a database
	query as follows: c.execute('INSERT INTO Table ({colnames}) VALUES
	{placeholders}', values). """
	col_names = ','.join(opt[0] for opt in options)
	placeholders = ','.join('?' for i in range(len(options)))
	if len(col_names) > 0:
		col_names = ',' + col_names
	if len(placeholders) > 0:
		placeholders = ',' + placeholders
	values = tuple(opt[1] for opt in options)
	return col_names, placeholders, values


def get_update_components(options):
	""" Same as get_insert_components but for update queries. Returns a tuple
	in the form (placeholders, values) to be used as follows:
	c.execute('UPDATE Table SET {placeholders}', values)"""
	placeholders = ','.join(
		'{}=?'.format(opt[0])
		for opt in options
	)
	values = tuple(opt[1] for opt in options)
	return placeholders, values


def check_options(options, allowed_options):
	for option, val in options:
		if option not in allowed_options:
			raise ValueError('Illegal option: {}'.format(option))


def rename_context(path, name):
	""" Returns a string which is what the path of the context would be if the
	context pointed to by `path` would be renamed with `name`."""
	return '.'.join(path.split('.')[:-1]+[name])


TASK_OPTIONS = {
	'title',
	'content',
	'created',
	'deadline',
	'start',
	'priority',
	'context',
	'done'
}

CONTEXT_OPTIONS = {
	'priority',
	'visibility'
}


class DataAccess():

	""" Wrap SQL operations into an methods-based interface. An instance of
	DataAccess is created thanks to a DB-API connection object that is
	connected to a sqlite database setup for todo.

	Some methods return instances of Row objects from the database. Such Row
	objects support the mapping protocol. More specifically, when the
	documentation mention Row-task objects, it's a mapping which represent a
	task with the following keys: id, title, created, deadline, start,
	priority, done, context, ctx_path where context is the context ID the task
	belongs to and ctx_path is the path of the same context.

	Row-context objects represent a context with the following keys: id, path,
	priority, visibility, own_tasks, total_tasks where own_tasks is the number
	of tasks which directly belong to the context and total_tasks is the
	number of tasks which belong to the context or one of the contexts in its
	descendance.

	The following list sums up the types and formats that must be used when
	calling methods and that are used in the returned values from the methods:

	 * Task ID are integers (type int). The hexadecimal representation is
	   dealt with by the calling application.

	 * Context path are fully dotted (non-root contexts start with a dot since
	   it's <empty string (name of the root context)> <dot> <name of the
	   subcontext>). The starting dot is removed by the calling application
	   for user convenience.

	 * Datetimes are strings in the %Y-%m-%d %H:%M:%S format (see Python
	   datetimes formatting reference)

	When a method accepts an `options` argument, it awaits a list of 2-tuple
	in the form (column name, value).
	"""

	def __init__(self, connection):
		self.connection = connection
		self.case_sensitive_like = False
		self.set_case_sensitive_like(True)
		c = self.connection.cursor()
		c.execute('PRAGMA foreign_keys = ON;')
		c.execute('PRAGMA synchronous = OFF;')
		self.connection.row_factory = sqlite3.Row
		self.changed_contexts = False

	def set_case_sensitive_like(self, switch=True):
		if self.case_sensitive_like == switch:
			return
		value = 'ON' if switch else 'OFF'
		c = self.connection.cursor()
		c.execute('PRAGMA case_sensitive_like = {};'.format(value))
		self.case_sensitive_like = switch

	def add_task(self, title, content, context='', options=[]):
		""" Add a task titled `title` and associated to the given `context`,
		with the given `options`. The context is created if not already
		existing."""
		check_options(options, TASK_OPTIONS)
		cid = self.get_or_create_context(context)
		query_tmp = """
			INSERT INTO Task (title, content, context {})
			VALUES (?, ?, ? {})
		"""
		col_names, placeholders, values = get_insert_components(options)
		values = (title, content, cid) + values
		query = query_tmp.format(col_names, placeholders)

		c = self.connection.cursor()
		c.execute(query, values)
		return c.lastrowid

	def update_task(self, tid, context=None, options=[]):
		""" Update the task identified by ID `tid` (int) with the given
		`options`. If the context of the task needs to be updated as well,
		then `context` should be passed the dotted path of the new context.
		Otherwise it should be None."""
		check_options(options, TASK_OPTIONS)
		if context is not None:
			cid = self.get_or_create_context(context)
			options = options.copy()
			options.append(('context', cid))
		query_tmp = """
			UPDATE Task SET {}
			WHERE id = ?
		"""
		placeholders, values = get_update_components(options)
		values += (tid,)
		query = query_tmp.format(placeholders)

		c = self.connection.cursor()
		c.execute(query, values)
		return c.rowcount

	def get_task(self, tid):
		""" Get the task identified by ID `tid` (int). Return a Row-Task object."""
		query = """
			SELECT t.*, c.path as ctx_path
			FROM Task t JOIN Context c
			ON t.context = c.id
			WHERE t.id = ?
		"""
		c = self.connection.cursor()
		c.execute(query, (tid,))
		row = c.fetchone()
		if row is None:
			return None
		return row

	def do_many(self, function, tids):
		""" Call the method `function` for each task ID in the `tids` list.
		`function` should accept only one positional argument (in addition to
		self) which is a task ID."""
		missing = []
		for tid in tids:
			updated = getattr(self, function)(tid)
			if updated == 0:
				missing.append(tid)
		return missing

	def set_done_many(self, tids):
		return self.do_many('set_done', tids)

	def remove_many(self, tids):
		return self.do_many('remove', tids)

	def set_done(self, tid):
		c = self.connection.cursor()
		c.execute("""
			UPDATE Task SET done = datetime('now')
			WHERE id = ?
			AND done IS NULL
		""", (tid,))
		return c.rowcount

	def remove(self, tid):
		c = self.connection.cursor()
		c.execute("""
			DELETE FROM Task
			WHERE id = ?
		""", (tid,))
		return c.rowcount

	def get_or_create_context(self, path, options=[]):
		""" Get the context whose path is `path`. If the context
		doesn't exists, it is created as well as well all necessary
		intermediary contexts. If the context is created, then `options` are
		applied to the newly created context.

		Return the ID of the context."""
		check_options(options, CONTEXT_OPTIONS)
		ctxs = path.split('.')[1:-1]
		path_so_far = ''
		c = self.connection.cursor()
		for ctx in ctxs:
			path_so_far += '.'+ctx
			try:
				c.execute("""
					INSERT INTO Context (path)
					VALUES (?)
				""", (path_so_far,))
			except sqlite3.IntegrityError: # Already exists
				continue
			else:
				self.changed_contexts = True

		query_tmp = """
			INSERT INTO Context (path {})
			VALUES (? {})
		"""
		col_names, placeholders, values = get_insert_components(options)
		query = query_tmp.format(col_names, placeholders)

		try:
			c.execute(query, (path,) + values)
		except sqlite3.IntegrityError as e:
			c2 = self.connection.cursor()
			c2.execute("""
				SELECT id FROM Context
				WHERE path = ?
			""", (path,))
			row = c2.fetchone()
			return row[0]
		else:
			self.changed_contexts = True
			return c.lastrowid

	def context_exists(self, path):
		""" Return a boolean indicating whether the context pointed to by the
		dotted path `path` exists."""
		c = self.connection.cursor()
		c.execute("""
			SELECT 1 FROM Context
			WHERE path = ?
		""", (path,))
		row = c.fetchone()
		return row is not None

	def get_basic_context_tally(self, path):
		""" Returns the number of (direct) tasks a context contains and the
		number of (direct subcontexts it contains, in the form of a
		2-tuple."""
		c = self.connection.cursor()
		c.execute("""
			SELECT COUNT(t.id)
			FROM Task t
			JOIN Context c
			  ON t.context = c.id
			WHERE t.done IS NULL
			  AND c.path = ?
			UNION ALL
			SELECT COUNT(id)
			FROM Context
			WHERE path LIKE ?
		""", (path, '{}_%'.format(path)))
		result = c.fetchall()
		return result[0][0], result[1][0]

	def set_context(self, path, options=[]):
		""" Set the context pointed to by `path` to have the given `options`.
		If the context doesn't already exist, it's created, which is why this
		method is called set_context and not update_context.

		Return the number of rows affected."""
		check_options(options, CONTEXT_OPTIONS)
		cid = self.get_or_create_context(path)
		query_tmp = """
			UPDATE Context SET {}
			WHERE id = ?
		"""
		placeholders, values = get_update_components(options)
		query = query_tmp.format(placeholders)

		c = self.connection.cursor()
		c.execute(query, values + (cid,))
		return c.rowcount

	def move(self, ctx1, ctx2):
		""" Move all the direct tasks from ctx1 to ctx2 (dotted paths).
		Doesn't affect subcontexts (and subtasks) of ctx1."""
		cid = self.get_or_create_context(ctx2)
		c = self.connection.cursor()
		c.execute("""
			UPDATE Task
			SET context = ?
			WHERE context = (
				SELECT id FROM Context
				WHERE path = ?
			)
		""", (cid, ctx1))

	def move_all(self, ctx1, ctx2):
		""" Same as `move` but move tasks of subcontexts as well. (any
		necessary context is created at the destination context."""
		for ctx in self.get_descendants(ctx1):
			dest = ctx2 + ctx['path'][len(ctx1):]
			self.move(ctx['path'], dest)

	def remove_context(self, path):
		""" Remove the context (and all subcontexts and tasks/subtasks)
		pointed to by `path`."""
		# We only have to remove the context and its subcontext as the foreign
		# key on tasks to set up to cascade the delete
		c = self.connection.cursor()
		c.execute("""
			DELETE FROM Context
			WHERE path LIKE ?
		""", ('{}%'.format(path),))
		self.changed_contexts = True
		return c.rowcount

	def rename_context(self, path, name):
		"""Rename context with given path with name. Returns None if new name
		already exists, number of row affected otherwise. `name` must NOT contain a dot.
		"""
		assert '.' not in name
		renamed = rename_context(path, name)
		# Lock the db at the first select to avoid race conditions
		self.connection.isolation_level = 'IMMEDIATE'
		# If the context (after renaming) already exists, abort
		c = self.connection.cursor()
		c.execute("""
			SELECT 1 FROM Context WHERE path = ?
		""", (renamed,))
		if c.fetchone() is not None:
			return None
		# We need to rename all the subcontexts as well. Can't do it in one
		# UPDATE query because sqlite's REPLACE would replace all occurrences
		# and we risk renaming other things (contexts whose same structure
		# repeats at a sub-level)
		c = self.connection.cursor()
		c.execute("""
			SELECT id, path FROM Context WHERE path LIKE ?
		""", ('{}%'.format(path),))
		new = ((renamed + row['path'][len(path):], row['id']) for row in c)
		c2 = self.connection.cursor()
		c2.executemany("""
			UPDATE Context
			SET path = ?
			WHERE ID = ?
		""", new)
		return c2.rowcount

	def todo(self, path='', recursive=False):
		""" Return a list of Row-tasks which belong the the context pointed to
		by `path`. If `recursive` is False, then the list only contains tasks
		that *directly* belong to the context. Otherwise it contains tasks
		from descendance as well. In the list, tasks are sorted by:
		  * priority, descending
		  * remaining time (before deadline, infinity if no deadline),
		    ascending
		  * datetime created, ascending
		"""
		if recursive:
			operator, value = 'LIKE', '{}%'.format(path)
		else:
			operator, value = '=', path
		c = self.connection.cursor()
		c.execute("""
			SELECT t.*, c.path as ctx_path
			FROM Task t JOIN Context c
			ON t.context = c.id
			WHERE c.path {} ?
			  AND t.done IS NULL
			  AND (c.path = ? OR c.visibility = 'normal')
			  AND (datetime('now')) >= datetime(t.start)
			ORDER BY
			  priority DESC,
			  COALESCE(
			      julianday(deadline),
			      julianday('9999-12-31 23:59:59')
			    ) - julianday('now') ASC,
			  created ASC
		""".format(operator), (value, path))
		return c.fetchall()

	def get_subcontexts(self, path='', get_empty=True):
		"""
		Return a list of Row-contexts that are direct children of the context
		pointed to by `path`. The list doesn't contain contexts that have a
		"hidden" visibility. If `get_empty` is False, then contexts that have
		0 total tasks are excluded from the list. Tasks in hidden contexts are
		not counted. In the list, contexts are sorted by:
		 * priority, descending
		 * total number of tasks (including tasks in descendance), ascending
		"""
		if get_empty:
			add_condition = ''
		else:
			add_condition = 'AND total_tasks > 0'
		c = self.connection.cursor()
		c.execute("""
			SELECT c.*, COUNT(own.id) as own_tasks, (
				SELECT COUNT(t.id)
				FROM Context c1
				LEFT JOIN Task t
				  ON t.context = c1.id
				 AND t.start <= (datetime('now'))
				 AND t.done IS NULL
				WHERE c1.path LIKE c.path||'%'
				  AND c1.visibility = 'normal'
			) as total_tasks
			FROM Context c
			LEFT JOIN Task own
			  ON own.context = c.id
			 AND own.start <= (datetime('now'))
			 AND own.done IS NULL
			WHERE path LIKE ?
			  AND path NOT LIKE ?
			  AND visibility = 'normal'
			  {}
			GROUP BY c.id
			ORDER BY
			  priority DESC,
			  total_tasks DESC
		""".format(add_condition), (
			'{}.%'.format(path),
			'{}.%.%'.format(path),
			)
		)
		return c.fetchall()

	def get_descendants(self, path=''):
		""" Return an iterator over Row-contexts that are in the descendance
		of the context pointed to by `path`. The iterator is agnostic of
		visibility and the contexts are sorted by their path. The first
		element of the iterator is the given context itself. """
		c = self.connection.cursor()
		c.execute("""
			SELECT c.*, COUNT(own.id) as own_tasks, (
				SELECT COUNT(t.id)
				FROM Context c1
				LEFT JOIN Task t
				  ON t.context = c1.id
				 AND t.start <= (datetime('now'))
				 AND t.done IS NULL
				WHERE c1.path LIKE c.path||'%' 
			) as total_tasks
			FROM Context c
			LEFT JOIN Task own
			  ON own.context = c.id
			 AND own.start <= (datetime('now'))
			 AND own.done IS NULL
			WHERE path LIKE ?
			GROUP BY c.id
			ORDER BY
			  c.path
		""", ('{}%'.format(path),))
		return c

	def history(self):
		""" Return an iterator over Row-tasks which iterates over all the
		tasks in existence, sorted by their date of creation."""
		c = self.connection.cursor()
		c.execute("""
			SELECT t.*, c.path as ctx_path
			FROM Task t JOIN Context c
			ON t.context = c.id
			ORDER BY t.created
		""")
		return c

	def get_greatest_id(self):
		""" Returns the greatest existing task ID, or None if there are no
		task."""
		c = self.connection.cursor()
		c.execute("""
			SELECT MAX(id)
			FROM Task
		""")
		row = c.fetchone()
		if row is None:
			return None
		else:
			return row[0]

	def purge(self, before):
		""" Remove all done tasks that were created before `before`. Remove
		all done tasks if `before` is None."""
		c = self.connection.cursor()
		query = """
			DELETE FROM Task
			WHERE done IS NOT NULL
		"""
		if before is not None:
			query += """
				AND created < ?
			"""
			values = (before,)
		else:
			values = ()
		c.execute(query, values)
		return c.rowcount

	def search(self, term, ctx='', done=None, before=None, after=None,
		       case=False):
		original = self.case_sensitive_like
		self.set_case_sensitive_like(case)
		c = self.connection.cursor()
		query = """
			SELECT t.*, c.path as ctx_path
			FROM Task t JOIN Context c
			ON t.context = c.id
			WHERE t.title LIKE ?
			  AND c.path LIKE ?
		"""
		params = ('%{}%'.format(term), '{}%'.format(ctx))

		if done is not None:
			cond = 'IS NOT NULL' if done else 'IS NULL'
			query += """
				AND t.done {}
			""".format(cond)
		if before is not None:
			query += """
				AND t.created < ?
			"""
			params = params + (before,)
		if after is not None:
			query += """
				AND t.created > ?
			"""
			params = params + (after,)

		c.execute(query, params)
		self.set_case_sensitive_like(original)
		return c.fetchall()

	def get_future_tasks(self):
		c = self.connection.cursor()
		now = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
		query = """
			SELECT t.*, c.path as ctx_path
			FROM Task t JOIN Context c
			ON t.context = c.id
			WHERE t.start > ?
			ORDER BY t.created
		"""
		c.execute(query, (now,))
		return c.fetchall()

	def take_editing_lock(self, tid):
		"""
		Set the column `editing` to 1 iff it's 0. Return True if the column was effectively set, False otherwise.
		"""
		c = self.connection.cursor()
		c.execute("""
			UPDATE Task
			SET editing = 1
			WHERE id = ?
			  AND editing = 0
		""", (tid,))
		taken = c.rowcount == 1
		self.connection.commit()
		return taken

	def release_editing_lock(self, tid):
		"""
		Release the editing lock on a task. The caller is trusted to have the
		lock and no verification is made.
		"""
		c = self.connection.cursor()
		c.execute("""
			UPDATE Task
			SET editing = 0
			WHERE id = ?
		""", (tid,))
		self.connection.commit()

	def exit(self, save=True):
		""" Close the database and save all operations done to it if `save` is
		True. Write all contexts paths (NON fully-dotted) to the contexts file
		if at least one context was created or removed during operations. The
		contexts file exists for terminal auto-completion."""
		if save:
			self.connection.commit()
			if self.changed_contexts:
				c = self.connection.cursor()
				c.execute("""
					SELECT DISTINCT path FROM Context
					ORDER BY path
				""")
				data_ctx = op.join(DATA_DIR, DATA_CTX_NAME)
				with open(data_ctx, 'w') as ctx_file:
					for row in c:
						ctx = userify_context(row[0])
						ctx_file.write(ctx + '\n')
		self.connection.close()



import textwrap, re


DEFAULT_WIDTH = 80


def wrap_text(text, width=DEFAULT_WIDTH, smart=False):
	wrapped_text = ""
	lines = text.splitlines()
	for i, line in enumerate(lines):
		if smart:
			go_wrap, sub_indent = smart_line(line)
		else:
			go_wrap, sub_indent = True, ''
		if go_wrap:
			prod_line = '\n'.join(textwrap.wrap(
				line,
				width=width,
				subsequent_indent=sub_indent
			))
		else:
			prod_line = line
		wrapped_text += prod_line
		if i < len(lines) - 1:
			wrapped_text += '\n'
	return wrapped_text


SPECIAL_LINES = [
	(re.compile('( {,3}> ?).*'),                   1), # Quote
	(re.compile('(#+ ?).*'),                       1), # Settext heading
	(re.compile('( {,3}([+\-*]|[0-9][.\)]) ?).*'), 1), # List item
]


def smart_line(line):
	if line.startswith(' '*4) or line.startswith('\t'): # Code block
		return False, ''
	for regex, gr in SPECIAL_LINES:
		match_obj = re.match(regex, line)
		if match_obj is not None:
			return True, ' '*len(match_obj.group(gr))
	return True, ''



import sqlite3

from . import utils


INIT_DB = [
	"""
	CREATE TABLE `Task` (
		`id`	INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE,
		`title`	TEXT NOT NULL,
		`created`	TEXT NOT NULL DEFAULT (datetime('now')),
		`deadline`	TEXT,
		`start`	TEXT NOT NULL DEFAULT (datetime('now')),
		`priority`	INTEGER NOT NULL DEFAULT 1,
		`done`	TEXT,
		`context`	INTEGER NOT NULL REFERENCES Context(id) ON DELETE CASCADE
	);
	""",
	"""
	CREATE TABLE `Context` (
		`id`	INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE,
		`path`	TEXT NOT NULL UNIQUE,
		`priority`	INTEGER NOT NULL DEFAULT 1,
		`visibility`	TEXT NOT NULL DEFAULT 'normal'
	);
	""",
	"""
	INSERT INTO `Context` (path) VALUES ('')
	""",
	"""
	CREATE INDEX `PathIndex` ON `Context` (`path` ASC);
	""",
	"""
	CREATE INDEX `DoneIndex` ON `Task` (`done`);
	""",
	"""
	CREATE INDEX `DateCreatedIndex` ON `Task` (`created`);
	""",
	"""
	ALTER TABLE Task ADD COLUMN `content` TEXT
	""",
	"""
	ALTER TABLE Task ADD COLUMN `editing` INTEGER NOT NULL DEFAULT 0
	"""
]


VERSIONS_INDEX = [
	('3.0', 0),
	('3.1', 4),
	('3.2', 6)
]


def update_database(path, current_version):
	if current_version is None:
		current_version = '0'
	for i, (version, idx) in enumerate(VERSIONS_INDEX):
		if utils.compare_versions(current_version, version) < 0:
			index = idx
			break
	else:
		index = len(INIT_DB)

	updates = INIT_DB[index:]
	if len(updates) > 0:
		conn = sqlite3.connect(path, isolation_level=None)
		for stmt in updates:
			conn.execute(stmt)
		conn.close()


def main():
	conn = sqlite3.connect('data.sqlite')
	for stmt in INIT_DB:
		conn.execute(stmt)


if __name__ == '__main__':
	main()



#! /usr/bin/env python3

import os, sys, sqlite3, functools, configparser, textwrap
import os.path as op
from datetime import date, datetime, timezone

from . import cli_parser, utils, data_access, core
from .bash_completion import installation as bash_completion_installation
from .rainbow import ColoredStr, cstr
from .data_access import DataAccess
from .utils import (
	DATA_DIR, DB_PATH, VERSION_PATH, DATAFILE_PATH, NOW,
	CannotOpenEditorError
)


__version__ = '3.4.2'


ISO_DATE_LENGTH = 10


# Icons used to print tasks' properties in the terminal.
# True is the ASCII version for challenged terminals.
# False is the Unicode version.
CONTEXT_ICON = {True: '#', False: '#'}
TIME_ICON = {True: '~', False: '⌛'}
PRIORITY_ICON = {True: '!', False: '★'}

WIDE_HIST_THRESHOLD = 120

TASK_MUTATORS = {
	'deadline': datetime.max,
	'start': None,
	'priority': 1,
	'title': None,
	'created': data_access.DATETIME_MIN
}

CONTEXT_MUTATORS = {
	'priority': 1,
	'visibility': 'normal'
}


CONFIG_FILE = op.expanduser(op.join('~', '.toduhrc'))

if os.name == 'posix':
	COLORS = 'on'
else:
	COLORS = 'off'

DEFAULT_CONFIG = configparser.ConfigParser()
DEFAULT_CONFIG['App'] = {
	'layout': 'basic',
	'todo_fashion': 'tidy',
	'show_empty_contexts': True
}
DEFAULT_CONFIG['Colors'] = {
	'colors': COLORS,
	'palette': '8',
	'id': 'yellow',
	'content': 'default',
	'context': 'cyan',
	'deadline': 'cyan',
	'priority': 'green',
	'done': 'green',
	'start': 'green',
}
DEFAULT_CONFIG['Word-wrapping'] = {
	'title': True,
	'content': True,
	'smart': False,
	'width': -1
}

CONFIG = configparser.ConfigParser(
	allow_no_value=True,
	strict=True
	)
# Loading the config with the default config
CONFIG.read_dict(DEFAULT_CONFIG)
# Loading the user config. Will complete/overwrite the default config
# but will keep default config entries that the user might have removed
CONFIG.read(CONFIG_FILE)

# Editor election: in config file? No -> in OS EDITOR variable? No -> vim
EDITOR = CONFIG.get('App', 'editor', fallback=None)
if EDITOR is None:
	EDITOR = os.environ.get('EDITOR', 'vim')

if CONFIG.getboolean('Colors', 'colors'):
	cstr = functools.partial(
		cstr,
		palette=CONFIG.get('Colors', 'palette')
	)
else:
	cstr = functools.partial(
		cstr,
		no_color=True
	)


DONE_STR = '[DONE]'


def main():
	argv = sys.argv[1:]
	if len(argv) == 1 and argv[0] == 'doduh':
		print('Beethoven - Symphony No. 5')
		sys.exit(0)
	args = cli_parser.parse_cli()

	if args.get('version'):
		print(__version__)
	elif args.get('location'):
		print(DATA_DIR)
	elif args.get('install_autocompletion'):
		bash_completion_installation.install_autocompletion()
	else:
		report = cli_parser.parse_args(args)
		if len(report) > 0:
			for error in report:
				print(error)
			sys.exit(1)

		current_version = get_installed_version()
		if not op.exists(DATA_DIR):
			os.mkdir(DATA_DIR)
		if current_version != __version__:
			with open(VERSION_PATH, 'w') as version_file:
				version_file.write(__version__)

		daccess = get_data_access(current_version)
		result = dispatch(args, daccess)
		if result is not None:
			feedback_code, *data = result
			globals()['feedback_'+feedback_code](*data)
		daccess.exit()


def get_installed_version():
	if op.exists(VERSION_PATH):
		with open(VERSION_PATH) as version_file:
			return version_file.read()
	else:
		if op.exists(DB_PATH):
			return '3.0.1'
		elif op.exists(DATAFILE_PATH):
			return '2.1'
		else:
			return None


def get_data_access(current_version):
	data_access.setup_data_access(current_version)
	connection = sqlite3.connect(DB_PATH)
	return DataAccess(connection)


# HANDLERS

# Do something with the args dictionary and the data access object then return
# a feedback code as well as some data about how things went. Feedback
# functions (whose names are feedback_<feedback_code>) will then be passed the
# data and should print some feedback based on the data. A handler can also
# return nothing (or None) if no feedback is intended.

def add_task(args, daccess):
	context = args.get('context')
	options = get_options(args, TASK_MUTATORS, {'deadline': {'None': None}})
	if context is None:
		context = ''

	if args['edit']:
		title, content = core.editor_edit_task(args['title'], None, EDITOR)
	else:
		title, content = args['title'], None

	id_ = daccess.add_task(title, content, context, options)
	return 'add_task', id_


def manage_task(args, daccess):
	tid = args['id'][0]
	context = args.get('context')
	options = get_options(args, TASK_MUTATORS, {'deadline': {'None': None}})

	if len(options) == 0 and context is None:
		return show_task(tid, daccess)
	else:
		upt_count = daccess.update_task(tid, context, options)
		if upt_count == 0:
			return 'task_not_found', tid


def show_task(tid, daccess):
	task = daccess.get_task(tid)
	if task is None:
		return 'task_not_found', tid
	# w3 = word-wrap width
	w3 = CONFIG.getboolean('Word-wrapping', 'content')
	if w3:
		w3 = CONFIG.getint('Word-wrapping', 'width')
		if w3 == -1:
			w3 = utils.get_terminal_width()
	else:
		w3 = None

	full_content = core.get_task_full_content(
		task['title'],
		task['content'],
		wrap_width=w3,
		smart_wrap=CONFIG.getboolean('Word-wrapping', 'smart')
	)

	return 'show_task', task, full_content


def edit_task(args, daccess):
	tid = args['id'][0]
	task = daccess.get_task(tid)
	if task is None:
		return 'task_not_found', tid
	can_edit = daccess.take_editing_lock(tid)
	if not can_edit:
		return 'cannot_edit', tid
	try:
		try:
			new_title, new_content = core.editor_edit_task(
				task['title'],
				task['content'],
				EDITOR
			)
		except CannotOpenEditorError as err:
			return 'cannot_open_editor', err.editor
		daccess.update_task(tid, options=[
			('title', new_title),
			('content', new_content)
		])
	finally:
		daccess.release_editing_lock(tid)


def do_task(args, daccess):
	not_found = daccess.set_done_many(args['id'])
	return 'multiple_tasks_done', not_found


def remove_task(args, daccess):
	not_found = daccess.remove_many(args['id'])
	return 'multiple_tasks_update', not_found


def manage_context(args, daccess):
	path = args['context']
	name = args.get('name')
	options = get_options(args, CONTEXT_MUTATORS)
	exists = True
	if len(options) == 0 and name is None:
		return todo(args, daccess)
	else:
		if name is not None:
			renamed = data_access.rename_context(args['context'], name)
			rcount = daccess.rename_context(args['context'], name)
			if rcount is None:
				return 'target_name_exists', renamed
			else:
				if rcount > 0:
					path = renamed
				else:
					exists = False
		if len(options) > 0:
			daccess.set_context(path, options)
		elif not exists:
			return 'not_exists', path


def move(args, daccess):
	ctx1 = args['ctx1']
	ctx2 = args['ctx2']
	source_exists = daccess.context_exists(ctx1)
	if not source_exists:
		return 'not_exists', ctx1
	else:
		daccess.move_all(ctx1, ctx2)


def remove_context(args, daccess):
	ctx = args['context']
	if not daccess.context_exists(ctx):
		return 'not_exists', ctx

	force = args['force']
	go_ahead = False
	if not force:
		nb_tasks, nb_subctx = daccess.get_basic_context_tally(ctx)
		ans = input('This context contains {} direct undone task(s) and '
			'{} subcontext(s). Continue? y/* '.format(nb_tasks, nb_subctx))
		go_ahead = ans == 'y'
	else:
		go_ahead = True
	if go_ahead:
		daccess.remove_context(ctx)


def todo(args, daccess):
	fashion = 'flat' if args['flat'] else None
	if fashion is None:
		fashion = 'tidy' if args['tidy'] else None
	if fashion is None:
		fashion = CONFIG.get('App', 'todo_fashion')
	ctx = args.get('context', '')
	if ctx is None:
		ctx = ''
	tasks = daccess.todo(ctx, recursive=(fashion == 'flat'))
	if fashion == 'tidy':
		get_empty = CONFIG.getboolean('App', 'show_empty_contexts')
		subcontexts = daccess.get_subcontexts(ctx, get_empty)
	else:
		subcontexts = []
	return 'todo', ctx, tasks, subcontexts


def get_contexts(args, daccess):
	path = args['context']
	if path is None:
		path = ''
	contexts = daccess.get_descendants(path)
	return 'contexts', contexts


def get_history(args, daccess):
	tasks = daccess.history()
	gid = daccess.get_greatest_id()
	return 'history', tasks, gid


def purge(args, daccess):
	force = args['force']
	before = args['before']
	go_ahead = False
	if force:
		go_ahead = True
	else:
		if before is None:
			q = "This will delete all done tasks. "
		else:
			q = "This will delete all done tasks created before "+\
			"{}. ".format(before)
		q += "This operation is irreversible. Continue? y/* "
		ans = input(q)
		if ans == 'y':
			go_ahead = True
	if go_ahead:
		count = daccess.purge(before)
		return 'purge', count


def search(args, daccess):
	term = args['term']
	done = None
	if args['done']:
		done = True
	elif args['undone']:
		done = False
	if args['context'] is None:
		ctx = ''
	else:
		ctx = args['context']
	tasks = daccess.search(
		term,
		ctx=ctx,
		done=done,
		before=args.get('before'),
		after=args.get('after'),
		case=args['case']
	)
	return 'todo', '', tasks, [], (term, args['case'])


def list_future_tasks(args, daccess):
	tasks = daccess.get_future_tasks()
	return 'todo', '', tasks, [], None


## DISPATCHING

# Map of the names of the commands to handlers defined above.

DISPATCHER = {
	'add': add_task,
	'task': manage_task,
	'edit': edit_task,
	'done': do_task,
	'rm': remove_task,
	'ctx': manage_context,
	'rmctx': remove_context,
	'mv': move,
	'contexts': get_contexts,
	'history': get_history,
	'purge': purge,
	'search': search,
	'future': list_future_tasks,
}


def dispatch(args, daccess):
	if 'command' in args:
		return DISPATCHER[args['command']](args, daccess)
	# If no command, fallback to the todo handler
	return todo(args, daccess)


def get_options(args, mutators, converters={}):
	"""
	Returns a list of 2-tuple in the form (option, value) for all options
	contained in the `mutators` collection if they're also keys of the `args`
	and have a non-None value. If the option (non- prefixed) is also a key of
	the `converters` dictionary then the associated value should be another
	dictionary indicating convertions to be done on the value found in `args`.
	e.g.
	args = {'deadline': 'none'}
	mutators = {'deadline'}
	converters = {'deadline': {'none': None}}
	=> [('deadline', None)]
	"""
	options = []
	for mutator in mutators:
		if mutator in args and args[mutator] is not None:
			val = args[mutator]
			convertions = converters.get(mutator)
			if convertions is not None and val in convertions:
				val = convertions[val]
			options.append((mutator, val))
	return options


## FEEDBACK FUNCTIONS

TASK_SUBCTX_SEP = '-'*40


def feedback_add_task(id_):
	pass


def feedback_task_not_found(tid):
	print('Task {} not found'.format(utils.to_hex(tid)))


def feedback_cannot_edit(tid):
	print('Task {} is already being edited'.format(utils.to_hex(tid)))


def feedback_cannot_open_editor(editor):
	print('Cannot open editor: {}'.format(editor))


def feedback_multiple_tasks_update(not_found):
	if len(not_found) > 0:
		s = 's' if len(not_found) > 1 else ''
		string = ', '.join(utils.to_hex(tid) for tid in not_found)
		print('Task{} not found: {}'.format(s, string))


def feedback_multiple_tasks_done(not_found):
	if len(not_found) > 0:
		string = ', '.join(utils.to_hex(tid) for tid in not_found)
		print('Not found or already done: {}'.format(string))


def feedback_todo(context, tasks, subcontexts, highlight=None):
	layout = CONFIG.get('App', 'layout')
	if layout == 'multiline':
		stringyfier = get_multiline_task_string
	else:
		stringyfier = get_basic_task_string

	if len(tasks) != 0:
		id_width = max(len(utils.to_hex(task['id'])) for task in tasks)
	else:
		id_width = 1

	for task in tasks:
		partial = functools.partial(stringyfier, context, id_width, task,
			highlight=highlight)
		safe_print(partial)
	if len(subcontexts) > 0:
		print(TASK_SUBCTX_SEP)
	for ctx in subcontexts:
		partial = functools.partial(get_context_string, context, id_width, ctx)
		safe_print(partial)


def feedback_target_name_exists(renamed):
	print('Context already exists: {}'.format(
		utils.get_relative_path('', renamed)
		)
	)


def feedback_not_exists(ctx):
	print("Context does not exist: {}".format(
		utils.get_relative_path('', ctx)
		)
	)


def feedback_contexts(contexts):
	def get_tally(ctx):
		return '{} ({})'.format(ctx['total_tasks'], ctx['own_tasks'])
	struct = [
		('context', lambda a: a, '<', 'path', lambda a: a[1:]),
		('visibility', 10, '<', 'visibility', None),
		('priority', 8, '<', 'priority', None),
		('undone tasks', 12, '<', None, get_tally)
	]
	utils.print_table(struct, contexts, is_context_default)


def feedback_history(tasks, gid):
	if gid is None:
		print('No history.')
	else:
		struct = get_history_struct(gid)
		utils.print_table(struct, tasks, is_task_default)


def feedback_purge(count):
	s = 's' if count > 1 else ''
	print('{} task{} deleted'.format(count, s))


def feedback_show_task(task, full_content):
	print(cstr("     ID:", '6'), utils.to_hex(task['id']))
	print(cstr("Created:", '6'), utils.sqlite_date_to_local(task['created']))
	if task['start'] == task['created']:
		print(cstr("  Start:", '6'), "@created")
	else:
		print(cstr("  Start:", '6'), utils.sqlite_date_to_local(task['start']))
	print(
		cstr(" Status:", '6'),
		"DONE" if task['done'] is not None else "TODO"
	)

	def print_metaline(ascii_):
		c = get_task_string_components(task, '', ascii_, highlight=None)
		if task['done'] is None:
			stuff = ['deadline', 'priority', 'context']
		else:
			stuff = ['priority', 'context']
		metaline = ' '.join(c[a] for a in stuff if c[a] != '')
		if len(metaline) > 0:
			return ' ' + metaline
		else:
			return None

	safe_print(print_metaline)

	print(cstr('-'*utils.get_terminal_width(), '3'))
	print(full_content)


# String building for todo feedback

# Those functions return a string. They accept a boolean `ascii_` argument
# that indicates whether to build the returned string with ASCII characters
# only (True) or whether non-ASCII characters are allowed (False). Those
# functions will then be partially called with all arguments set except the
# `ascii_` one and the resulting partial will be passed to `safe_print` which
# will take care of trying to print the non-ASCII version and then fallback to
# the ASCII version in case of error from the terminal.

def get_basic_task_string(context, id_width, task, highlight=None, ascii_=False):
	c = get_task_string_components(task, context, ascii_, highlight=highlight)

	if isinstance(c['id'], ColoredStr):
		ansi_offset = c['id'].lenesc
	else:
		ansi_offset = 0
	result = ' {id:>{width}} | '.format(width=id_width + ansi_offset, **c)
	left_width = id_width + 4
	init_indent = left_width

	if c['done']:
		adding = c['done'] + ' '
		result += adding
		init_indent += len(DONE_STR) + 1 # [DONE] followed by space

	wrap_width = CONFIG.getint('Word-wrapping', 'width')
	if wrap_width == -1:
		wrap_width = utils.get_terminal_width()

	if CONFIG.getboolean('Word-wrapping', 'title'):
		title_subindent = ' '*left_width

		# The correct way to wrap would be to order textwrap to wrap the whole
		# ` {id} | {title}` with the subsequent indent being the length of `
		# {id} | `. However, {id} containing ANSI escape characters for
		# coloring will mess up textwrap character counting, so what we do
		# instead is wrapping only {title} prefixed with the length of ` {id}
		# | `, and we remove the prefix afterwards, ` {id} | ` taking its
		# place.

		lines = textwrap.wrap(
			' '*init_indent + c['title'],
			width=wrap_width,
			subsequent_indent=' '*left_width
		)
		lines[0] = lines[0][init_indent:]
	else:
		lines = [c['title']]

	len_last_line = len(lines[-1])
	title = '\n'.join(lines)

	start_title = len(result)
	result += title
	end_title = len(result)

	metadata = [c['deadline'], c['priority'], c['context'], c['start']]
	metatext = ' '.join(stuff for stuff in metadata if stuff != '')

	if len(metatext) > 0:
		wrap_title = CONFIG.getboolean('Word-wrapping', 'title')
		not_enough_space = wrap_width - len_last_line <= 0
		if wrap_title and not_enough_space:
			result += '\n' + ' '*left_width
		else:
			result += ' '
		result += metatext

	return result


def get_multiline_task_string(context, id_width, task, highlight=None, ascii_=False):
	c = get_task_string_components(task, context, ascii_, highlight=highlight)
	template = ' {id} {done} / {deadline} {priority} {context} {start}\n'
	result =  template.format(**c)

	wrap_width = CONFIG.getint('Word-wrapping', 'width')
	if wrap_width == -1:
		wrap_width = utils.get_terminal_width()
	title = c['title']
	if CONFIG.getboolean('Word-wrapping', 'title'):
		title = '\n'.join(textwrap.wrap(c['title'], width=wrap_width))
	result += title + '\n'
	return result


def get_task_string_components(task, ctx, ascii_=False, highlight=None):
	id_str = cstr(utils.to_hex(task['id']), clr('id'))

	if highlight is not None and CONFIG.getboolean('Colors', 'colors'):
		term, case = highlight
		content_str = utils.get_highlights_term(
			task['title'],
			term,
			(clr('content'), CONFIG.get('Colors', 'palette')),
			case=case
		)
	else:
		content_str = cstr(task['title'], clr('content'))

	remaining_str = ''
	deadline = get_datetime(task['deadline'])
	if deadline is not None:
		remaining = deadline - NOW
		user_friendly = utils.parse_remaining(remaining)
		remaining_str = '{} {} remaining'.format(
			TIME_ICON[ascii_],
			user_friendly
		)
		remaining_str = cstr(remaining_str, clr('deadline'))

	prio_str = ''
	priority = task['priority']
	if not is_task_default(task, 'priority'):
		prio_str = '{}{}'.format(PRIORITY_ICON[ascii_], priority)
		prio_str = cstr(prio_str, clr('priority'))

	ctx_path = utils.get_relative_path(ctx, task['ctx_path'])
	if ctx_path == '':
		ctx_str = ''
	else:
		ctx_str = '{}{}'.format(CONTEXT_ICON[ascii_], ctx_path)
		ctx_str = cstr(ctx_str, clr('context'))

	done_str = ''
	if task['done'] is not None:
		done_str = cstr(DONE_STR, clr('done'))

	start_str = ''
	start_date = utils.sqlite_date_to_local(task['start'])[:ISO_DATE_LENGTH]
	if start_date > date.today().isoformat():
		start_str = cstr('[starts: {}]'.format(start_date), clr('start'))

	return {
		'id': id_str,
		'title': content_str,
		'deadline': remaining_str,
		'priority': prio_str,
		'context': ctx_str,
		'done': done_str,
		'start': start_str,
	}


def get_context_string(context, id_width, ctx, ascii_=False):
	hash_str = cstr('#', clr('id'))
	if isinstance(hash_str, ColoredStr):
		ansi_offset = hash_str.lenesc
	else:
		ansi_offset = 0
	path = utils.get_relative_path(context, ctx['path'])
	string = '{hash:>{width}} | {path} ({nbr})'.format(
		hash=hash_str,
		width=id_width + ansi_offset + 1,
		path=path,
		nbr=ctx['total_tasks']
	)
	priority = ctx['priority']
	if not is_task_default(ctx, 'priority'):
		prio_str = ' {}{}'.format(PRIORITY_ICON[ascii_], priority)
		string += cstr(prio_str, clr('priority'))
	return string


def safe_print(partial):
	try:
		result = partial(ascii_=False)
		if result is not None:
			print(result)
	except UnicodeEncodeError:
		result = partial(ascii_=True)
		if result is not None:
			print(result)


def get_datetime(db_dt):
	""" Get a datetime object from the string retrieved from the database."""
	if db_dt is None:
		return None
	return datetime\
		.strptime(db_dt, utils.SQLITE_DT_FORMAT)\
		.replace(tzinfo=timezone.utc)


def is_task_default(task, prop):
	if prop == 'start':
		return task['start'] == task['created']
	return is_default(task, prop, TASK_MUTATORS)


def is_context_default(ctx, prop):
	return is_default(ctx, prop, CONTEXT_MUTATORS)


def is_default(dico, prop, defaults):
	default = defaults.get(prop, None)
	if default is None:
		return False
	else:
		return dico[prop] == default


def get_history_struct(gid):
	gid_len = len(utils.to_hex(gid))
	struct = [
		('id', gid_len + 1, '>', 'id', utils.to_hex),
		('title', lambda a: 3 * (a//4), '<', 'title', None),
		('created', 19, '<', 'created', utils.sqlite_date_to_local),
	]
	if utils.get_terminal_width() > WIDE_HIST_THRESHOLD:
		struct += [
			('start', 19, '<', 'start', utils.sqlite_date_to_local),
			('deadline', 19, '<', 'deadline', utils.sqlite_date_to_local),
			('priority', 8, '>', 'priority', None)
		]
	struct += [
		('context', lambda a: a//4 + a%4, '<', 'ctx_path', lambda a: a[1:]),
		('status', 7, '<', 'done', lambda a: 'DONE' if a is not None else '')
	]
	return struct


def clr(component):
	return CONFIG.get('Colors', component)



from . import utils, text_wrap


def editor_edit_task(title, content, editor):
	"""
	Opens the text editor `editor` to edit a task's `title` and `content`.
	Returns the updated title and content after editing is done.
	"""
	init_content = get_task_full_content(title, content)
	full_content = utils.input_from_editor(init_content, editor)
	title, content = parse_task_full_content(full_content)
	return title, content


def get_task_full_content(title, content, wrap_width=None, smart_wrap=False):
	"""
	Return the full text of a task from its `title` and `content`. If
	`wrap_width` is not None, then both the title and content are word-wrapped
	to be contained in the given width. If `smart_wrap` is True, then the
	word-wrapping cleverly handles things such as list, quotes, etc.
	"""
	if wrap_width is not None:
		title = text_wrap.wrap_text(title, wrap_width, smart_wrap)
		if content is not None:
			content = text_wrap.wrap_text(content, wrap_width, smart_wrap)
	title_width = max(len(line) for line in title.splitlines())
	if content is None:
		return title
	else:
		return '{}\n{}\n{}'.format(title, '='*title_width, content)


def parse_task_full_content(full_content):
	"""
	Return a tuple (title, content) extracted from the content found in a file
	edited through `todo edit` or `todo add [<title>] --edit`
	"""
	title, content = '', None
	state = 'title'
	lines = full_content.splitlines(keepends=True)
	for i, line in enumerate(lines):
		if state == 'title' and line.startswith('==='):
			state = 'content'
			continue
		if state == 'title':
			title += line
		elif state == 'content':
			if content is None:
				content = ''
			content += line

	# Removes blank characters at the right of the title (newline leading to
	# settext heading underlining and potential newline added by text editors)
	title = title.rstrip()

	return title, content



import sys, argparse, functools

from . import data_access, utils
from .utils import NOW


COMMANDS = {
	'add', 'done', 'task', 'edit', 'rm', 'ctx', 'contexts', 'history',
	'purge', 'mv', 'rmctx', 'search', 'future', '-h', '--help', '--location',
	'--version', '--install-autocompletion'}


## Argument parsing error messages

INCORRECT_PRIORITY = 'PRIORITY must be an integer.'
INCORRECT_VISIBILITY = "VISIBILITY must be 'normal' or 'hidden'."
INCORRECT_MOMENT = "MOMENT must be in the YYYY-MM-DD format, or the "+\
                   "YYYY-MM-DD HH:MM:SS format, or a delay in the "+\
                   "([0-9]+)([wdhms]) format."
INCORRECT_CTX_RENAME = "Can't use '.' in context new name "+\
                       "(only right-most context name is updated)."
CANT_RENAME_ROOT = "Can't rename root context."
INVALID_TID = "Invalid task{} ID: {}"


# ARGUMENT PARSERS.
#
# Each function should return a 2-tuple where the first component is a boolean
# indicating whether the value of the argument is correct and has been
# successfully parsed. The second component contains either the parsed value
# (if success) or an error message.

def parse_id(tid_list):
	if isinstance(tid_list, str):
		# In the case of the task command, we actually get a lone ID
		tid_list = [tid_list]
	valid = []
	invalid = []
	for tid in tid_list:
		try:
			valid.append(int(tid, 16))
		except ValueError:
			invalid.append(tid)
	if len(invalid) == 0:
		return True, valid
	else:
		s = 's' if len(invalid) > 1 else ''
		string = ', '.join(invalid)
		error = INVALID_TID.format(s, string)
		return False, error


def parse_context(ctx):
	return True, data_access.dbfy_context(ctx)


def parse_moment(moment, direction=1):
	""" Parse a moment, which can be either a string datetime in the allowed
	datetimes format, either a delay (e.g. 2w). In the case of a delay,
	direction indicates in which direction in time the delay is applied to the
	current time. It can either be 1 (future) or -1 (past)."""
	dt = utils.get_datetime(moment, NOW, direction)
	if dt is None:
		return False, INCORRECT_MOMENT
	else:
		return True, dt.strftime(utils.SQLITE_DT_FORMAT)


def parse_deadline(moment):
	""" A deadline-specific wrapper around parse_moment. Case-insensitive
	'none' is accepted and is parsed as 'None' (the string)"""
	# The reason why it returns the string 'None' and not the value None is
	# that docopt gives the value None to all arguments and options that
	# weren't used. Using 'None' (the string) allows us the make the difference
	# between a deadline set as none and no deadline set.
	if moment.lower() == 'none':
		return True, 'None'
	else:
		return parse_moment(moment)


def parse_new_context_name(name):
	if '.' in name:
		return False, INCORRECT_CTX_RENAME
	elif name == '':
		return False, CANT_RENAME_ROOT
	else:
		return True, name


PARSERS = [
	('id', parse_id),
	('context', parse_context),
	('ctx1', parse_context),
	('ctx2', parse_context),
	('deadline', parse_deadline),
	('start', parse_moment),
	('before', functools.partial(parse_moment, direction=-1)),
	('after', functools.partial(parse_moment, direction=-1)),
	('name', parse_new_context_name)
]


def parse_args(args):
	""" Apply application-level parsing of the values of the args dictionary
	*in place*. Returns a report which is a list of errors (strings) that
	might have occured during parsing. There's no waranty that the args
	dictionary will work with the rest of the application if the report
	list isn't empty."""
	report = []
	for arg_name, parser in PARSERS:
		value = args.get(arg_name)
		if value is not None:
			success, result = parser(value)
			if success:
				args[arg_name] = result
			else:
				report.append(result)
	return report


def parse_cli():
	argv = sys.argv[1:]
	if len(argv) == 0:
		argv = [''] # bare todo with root context
	command, params = argv[0], argv[1:]
	if command in COMMANDS:
		return parse_command(argv)
	else:
		return parse_bare_todo(argv)


def parse_bare_todo(argv):
	parser = argparse.ArgumentParser()
	parser.add_argument('context', nargs='?',
		help="Context to show the tasks of. Defaults to the root context "
		     "(represented by an empty string)"
	)

	style_group = parser.add_mutually_exclusive_group()
	style_group.add_argument('--flat', action='store_true',
		help="Show the tasks of subcontexts as well"
	)
	style_group.add_argument('--tidy', action='store_true',
		help="Only show the tasks of the given context, and list subcontexts"
	)

	args = parser.parse_args(argv)
	return vars(args)


def parse_command(argv):
	parser = argparse.ArgumentParser(
		description="Alternatively, the program might be called this way:\n"
		    "  $ todo [<context>]\n"
		    "to show the tasks of a specific context. If <context> is "
		    "omitted, then the root context is used.",
		formatter_class=argparse.RawTextHelpFormatter
	)
	root_group = parser.add_mutually_exclusive_group()
	root_group.add_argument('--location', action='store_true',
		help="Print the location of the todo data directory"
	)
	root_group.add_argument('--version', action='store_true',
		help="Print the version of the program"
	)
	root_group.add_argument('--install-autocompletion', action='store_true',
		help="Install command-line autocompletion for todo",
	)

	subparsers = parser.add_subparsers()

	add_parser = subparsers.add_parser('add',
		help="Add a new task")
	add_parser.set_defaults(command='add')
	add_parser.add_argument('title',
		help="The title of the task"
	)
	add_parser.add_argument('-d', '--deadline',
		help="Deadline of the task, in the YYYY-MM-DD (ISO 8601) format, or "
		     "as a delay in the <n>(s|m|h|d|w) format, where <n> is an "
		     "integer and what follows represents respectively seconds, "
		     "minutes, hours, days or weeks"
	)
	add_parser.add_argument('-s', '--start',
		help="Start line of the task, in the same format than --deadline. "
		     "Defaults to the moment the task is created."
	)
	add_parser.add_argument('-c', '--context',
		help="Context to put the task in. Defaults to the root context"
	)
	add_parser.add_argument('-p', '--priority', type=int,
		help="Priority of the task, as an integer. Higher the interger, "
		     "higher the priority"
	)
	add_parser.add_argument('-e', '--edit', action='store_true',
		help="Edit the task in a text editor before adding it"
	)

	search_parser = subparsers.add_parser('search',
		help="Search for tasks")
	search_parser.set_defaults(command='search')
	search_parser.add_argument('term',
		help="Substring to be searched in tasks' titles."
	)
	search_parser.add_argument('-c', '--context',
		help="Context to search in, with recursion"
	)
	search_parser.add_argument('--before',
		help="Restrict the search to tasks created before a certain moment. "
		     "Same format than <add --deadline> except that a delay is a "
		     "delay in the past."
	)
	search_parser.add_argument('--after',
		help="Restrict the search to tasks created after a certain moment. "
		     "Same format than <add --deadline>"
	)
	search_parser.add_argument('--case', action='store_true',
		help="Make the search case-sensitive"
	)
	done_group = search_parser.add_mutually_exclusive_group()
	done_group.add_argument('--done', action='store_true',
		help="Restrict the search to done tasks"
	)
	done_group.add_argument('--undone', action='store_true',
		help="Restrict the search to undone tasks"
	)

	done_parser = subparsers.add_parser('done',
		help="Set task(s) as done")
	done_parser.set_defaults(command='done')
	done_parser.add_argument('id', nargs='+',
		help="The list of tasks' IDs to set as done"
	)

	task_parser = subparsers.add_parser('task',
		help="Select a task to apply modifiers to. Options that are shared "
		     "with the add command are documented there")
	task_parser.set_defaults(command='task')
	task_parser.add_argument('id',
		help="ID of the task to apply modifiers to"
	)
	task_parser.add_argument('-d', '--deadline')
	task_parser.add_argument('-s', '--start')
	task_parser.add_argument('-c', '--context')
	task_parser.add_argument('-p', '--priority')
	task_parser.add_argument('-t', '--title',
		help="Rename the task"
	)

	edit_parser = subparsers.add_parser('edit',
		help="Edit the title of a task")
	edit_parser.set_defaults(command='edit')
	edit_parser.add_argument('id',
		help="ID of the task to edit"
	)

	rm_parser = subparsers.add_parser('rm',
		help="Remove task(s) from history")
	rm_parser.set_defaults(command='rm')
	rm_parser.add_argument('id', nargs='+',
		help="The list of tasks' IDs to remove"
	)

	ctx_parser = subparsers.add_parser('ctx',
		help="Select a context to show the tasks of, or to apply modifiers to")
	ctx_parser.set_defaults(command='ctx')
	ctx_parser.add_argument('context',
		help="Context to show or to modify"
	)
	fashion_group = ctx_parser.add_mutually_exclusive_group()
	fashion_group.add_argument('--flat', action='store_true',
		help="Show the tasks of subcontexts as well"
	)
	fashion_group.add_argument('--tidy', action='store_true',
		help="Only show the tasks of the given context, and list subcontexts"
	)
	ctx_parser.add_argument('-p', '--priority', type=int,
		help="The priority of the context, as an integer. The higher the "
		     "integer, the higher the priority. Contexts with a higher "
		     "priority show up first in the list of sub-contexts of a "
		     "context in a todolist"
	)
	ctx_parser.add_argument('-v', '--visibility', choices=['normal', 'hidden'],
		help="Hidden contexts don't show up in the list of sub-contexts of a "
		     "context in a todolist"
	)
	ctx_parser.add_argument('--name',
		help="Name of a context. Cannot contain a dot."
	)

	mv_parser = subparsers.add_parser('mv',
		help="Move all tasks and subcontexts from one context to another")
	mv_parser.set_defaults(command='mv')
	mv_parser.add_argument('ctx1',
		help="Source context"
	)
	mv_parser.add_argument('ctx2',
		help="Destination context"
	)

	rmctx_parser = subparsers.add_parser('rmctx',
		help="Remove a context and all its tasks")
	rmctx_parser.set_defaults(command='rmctx')
	rmctx_parser.add_argument('context',
		help="Context to remove"
	)
	rmctx_parser.add_argument('--force', action='store_true',
		help="Removing a context requires user interaction to confirm the "
		     "action. Setting this option to true skips this step."
	)

	contexts_parser = subparsers.add_parser('context',
		help="List all contexts")
	contexts_parser.set_defaults(command='context')
	contexts_parser.add_argument('context',
		help="Restrict the list to subcontexts of the given context"
	)

	history_parser = subparsers.add_parser('history',
		help="Show tasks history")
	history_parser.set_defaults(command='history')

	purge_parser = subparsers.add_parser('purge',
		help="Remove done tasks from history")
	purge_parser.set_defaults(command='purge')
	purge_parser.add_argument('--force', action='store_true',
		help="Purging requires user interaction to confirm the "
		     "action. Setting this option to true skips this step."
	)
	purge_parser.add_argument('--before',
		help="Only remove done tasks that were created before the given "
		     "moment. Same format than <search --before>"
	)

	future_parser = subparsers.add_parser('future',
		help="Show tasks that will start in the future"
	)
	future_parser.set_defaults(command='future')

	return vars(parser.parse_args(argv))



import re, os
import os.path as op
from datetime import datetime, timedelta, timezone

from . import rainbow


DATA_DIR_NAME = '.toduh'
DATAFILE_NAME = 'data.json'
DATABASE_NAME = 'data.sqlite'
DATA_CTX_NAME = 'contexts'
VER_FILE_NAME = 'version'

# If a .toduh exists in the current working directory, it's used by the
# program. Otherwise the one in the home is used.
if op.exists(DATA_DIR_NAME) and op.isdir(DATA_DIR_NAME):
	DATA_DIR = DATA_DIR_NAME
else:
	DATA_DIR = op.expanduser(op.join('~', '.toduh'))

DB_PATH = op.join(DATA_DIR, DATABASE_NAME)
VERSION_PATH = op.join(DATA_DIR, VER_FILE_NAME)
DATAFILE_PATH = op.join(DATA_DIR, DATAFILE_NAME)


ISO_SHORT = '%Y-%m-%d'
ISO_DATE = ISO_SHORT+'T%H:%M:%SZ'
USER_DATE_FORMATS = [
	ISO_SHORT,
	ISO_SHORT+'T%H:%M:%S',
	ISO_SHORT+' %H:%M:%S'
]

REMAINING = {
	'w': 7*24*3600,
	'd': 24*3600,
	'h': 3600,
	'm': 60,
	's': 1
}
REMAINING_RE = re.compile('\A([0-9]+)([wdhms])\Z')

SQLITE_DT_FORMAT = '%Y-%m-%d %H:%M:%S'

NOW = datetime.utcnow().replace(tzinfo=timezone.utc)


def print_table(struct, iterable, is_default=lambda obj, p: False):
	""" This function, which is responsible for printing tables to the
	terminal, awaits a "structure", an iterable and a function. The structure
	describes the columns of the table and their properties. It's a list of
	tuples where each tuple describes one column of the table. A tuple has 5
	elements corresponding to the following pieces of information:
	 1. The header of the column, a string
	 2. The width of the column given in number of characters. The width can
	    either be an integer or a function accepting one argument. Widths
	    given as integers will be subtracted from the terminal's width to
	    obtain the "available space". After that, widths given as functions
	    will be evaluated with the available space given as their argument and
	    the functions should return an integer being the actual width of the
	    corresponding column.
	 3. How the name of the column should be aligned in the table header.
	    Value should be either ">", "<", "=" or "^". See Python's format mini-
	    language.
	 4. For mappings, the name of the key of the map to print the value of. If
	    this element is set to None, the object itself will be used for
	    printing.
	 5. A function which takes as argument the value obtained according to the
	    previous element and return the string to finally print.

    The function `is_default` should accept a yielded object and the element 4
    of the tuple and returns True if this objects contains a "default value"
    at the given key. Such values aren't printed in the table.

    See the function get_history_struct to have an example of structure."""
	term_width = get_terminal_width()
	occupied = sum(w if isinstance(w, int) else 0 for _, w, *_ in struct)
	available = term_width - occupied - (len(struct) - 1)
	template, separator = '', ''
	widths = {}
	for header, width, align, *_ in struct:
		w = max(0, width if isinstance(width, int) else width(available))
		widths[header] = w
		template = ' '.join([template, '{{: {}{}}}'.format(align, w)])
		separator = ' '.join([separator, '-'*w])
	template, separator = template[1:], separator[1:] # Starting space

	table = template.format(*(t[0] for t in struct))
	table = '\n'.join([table, separator])
	for obj in iterable:
		values = []
		for h, _, _, a, f in struct:
			f = f if f is not None else lambda a: a
			if is_default(obj, a):
				value = ''
			else:
				if a is None:
					value = obj
				else:
					value = obj[a]
				value = f(value)
			if value is None:
				value = ''
			value = str(value).split('\n')[0]
			value = limit_str(str(value), widths[h])
			values.append(value)
		line = template.format(*values)
		table = '\n'.join([table, line])
	print(table)


def limit_str(string, length):
	if len(string) <= length:
		return string
	else:
		if length <= 3:
			return string[:length]
		else:
			return string[:length-3] + '...'


def get_datetime(string, now, direction=1):
	"""Parse the string `string` representating a datetime. The string can be
	a delay such `2w` which means "two weeks". In this case, the datetime is
	the datetime `now` plus/minus the delay. The `direction` option indicates
	if the delay needs to be added to now (+1) or substracted from now (-1).
	In any case, this returns a datetime object."""
	match = REMAINING_RE.match(string)
	if match is not None:
		value, unit = match.groups()
		seconds = int(value) * REMAINING[unit]
		offset = direction * timedelta(seconds=seconds)
		return now + offset
	else:
		dt = None
		for pattern in USER_DATE_FORMATS:
			try:
				dt = datetime.strptime(string, pattern)
			except ValueError:
				continue
			else:
				dt = datetime.utcfromtimestamp(dt.timestamp())
				dt = dt.replace(tzinfo=timezone.utc)
				break
		return dt


def parse_remaining(delta):
	null_delta = timedelta(0)
	sign = '' if delta >= null_delta else '-'
	delta = abs(delta)
	seconds = 3600 * 24 * delta.days + delta.seconds
	if seconds >= 2 * 24 * 3600:
		string = '{} days'.format(delta.days)
	elif seconds >= 2*3600:
		string = '{} hours'.format(24*delta.days + delta.seconds // 3600)
	elif seconds >= 2*60:
		string = '{} minutes'.format(seconds // 60)
	else:
		string = '{} seconds'.format(seconds)
	return '{}{}'.format(sign, string)


def input_from_editor(init_content, editor):
	import subprocess
	with CustomTemporaryFile() as filename:
		with open(filename, 'w') as edit_file:
			edit_file.write(init_content)
		try:
			subprocess.call([editor, filename])
		except FileNotFoundError:
			raise CannotOpenEditorError('Cannot open editor', editor=editor)
		with open(filename) as edit_file:
			new_content = edit_file.read()
	return new_content


class CannotOpenEditorError(Exception):

	def __init__(self, message, editor):
		super().__init__(message)
		self.editor = editor


class CustomTemporaryFile:

	def __enter__(self):
		import uuid
		self.path = op.join(DATA_DIR, '.todoedit-'+uuid.uuid4().hex)
		return self.path

	def __exit__(self, type_, value, traceback):
		os.remove(self.path)
		return type_ is None


def get_relative_path(parent, desc):
	rel = desc[len(parent):]
	if rel.startswith('.'):
		rel = rel[1:]
	return rel


def to_hex(integer):
	return hex(integer)[2:] # 0x...


def get_terminal_width():
	try:
		size = os.get_terminal_size()[0]
	except OSError:
		size = 80
	return size


def sqlite_date_to_local(sqlite_date):
	if sqlite_date is None:
		return ''
	try:
		dt = datetime.strptime(sqlite_date, SQLITE_DT_FORMAT)
	except ValueError:
		return ''
	dt = dt.replace(tzinfo=timezone.utc)
	try:
		local_dt = dt.astimezone(tz=None)
	except ValueError:
		# Some exotic dates such as '0001-01-01 00:00:00' don't work well
		return sqlite_date
	return local_dt.strftime(SQLITE_DT_FORMAT)


def get_highlights_term(string, term, str_color, case=False):
	if str_color is not None:
		escape = rainbow.get_escape(*str_color)
		if escape is None:
			str_color = None

	if len(term) > 0:
		def term_repl(matchobj):
			repl = '\33[1;31m' + matchobj.group(0) + '\33[0m'
			if str_color is not None:
				repl = '\33[0m' + repl + escape
			return repl
		args = dict(
			pattern=term,
			repl=term_repl,
			string=string
		)
		if not case:
			args.update(flags=re.IGNORECASE)
		highlighted = re.sub(**args)
		if str_color:
			highlighted = escape + highlighted + '\33[0m'
		return highlighted
	else:
		if str_color is not None:
			return escape + string + '\33[0m'
		else:
			return string


def compare_versions(vA, vB):
	# ((major, minor, patch), tag, tagNumber)
	(releaseA, tagA, tagnumA), (releaseB, tagB, tagnumB) = [
		parse_version(v) for v in [vA, vB]
	]
	if releaseA != releaseB:
		return 1 if releaseA > releaseB else -1
	else:
		if tagA is None:
			if tagB is None:
				return 0
			elif tagB < 'final':
				return 1
			else:
				return -1
		elif tagB is None:
			if tagA < 'final':
				return -1
			else:
				return 1
		elif tagA != tagB:
			return 1 if tagA > tagB else -1
		else:
			if tagnumA > tagnumB:
				return 1
			elif tagnumA == tagnumB:
				return 0
			else:
				return -1


def parse_version(v):
	"""
	A version is in the form <major>.<minor>.<patch><tag><tagnumber>
	<minor> and <patch> can be omitted (in which case they count for 0)
	<tag> and <tagnumber> are optional
	"""
	undotted = v.split('.')
	if len(undotted) == 0:
		raise ValueError("Versio number cannot be empty")
	if len(undotted) > 3:
		raise ValueError("Version number cannot have more than 3 dots")
	tag_match = re.match('([0-9]+)([a-z]+)([0-9]+)?', undotted[-1])
	if tag_match is not None:
		least_number, tag, tagnumber = tag_match.groups()
	else:
		least_number, tag, tagnumber = undotted[-1], None, None
	if tagnumber is None:
		tagnumber = 0
	release = tuple(undotted[:-1]) + (least_number,)
	while len(release) < 3:
		release = release + (0,)
	release = tuple(int(n) for n in release)
	return (release, tag, int(tagnumber))



""" This module deals with support of the color in the terminal through ANSI
escape codes. It defines the high level class ColoredStr which is a string
containing color escape codes ready to be printed in color to the terminal.
"""

import re


# There are three possible types of ANSI escape codes to print text in colors:
# - the first one accepts an integer in the range [0;7] which indicates one of
#   the primary colors among black, red, green, yellow, blue, magenta, cyan,
#   and white, in order. This escape code can actually be mixed with another
#   one that put the text in bold and which has the side effect of making the
#   color brighter (the normal 8 colors are quite dark). I chose not to deal
#   with this second code.
# - the second one accepts an integer in the range [0;255] indicating a color
#   among a specific palette of 256 colors. In the UNIX world, most terminals
#   work with the x-term palette. Such a palette follows a specific pattern
#   and conversions between RGB and code-point in the palette can be computed
#   without carying a whole lookup table.
# - the third one accepts three integers in the range [0;255] defining the
#   color in RGB coordinates.

# Support of these codes varies from one terminal to another. Most program
# that prints stuff in color stick to the 8 basic colors, which are widely
# supported. I prefer to let the user chose what he wants to use via a
# configuration file. Using an unsupported escape code can result in no color
# at all, a wrong color, or the escape code being printed literally.

ANSI_TEMPLATES = {
	'8':         '\33[3{}m',          # Basic 8 colors
	'xterm-256': '\33[38;5;{}m',      # 256 colors
	'rgb':       '\33[38;2;{};{};{}m' # 24-bit RGB true colors
}

# ANSI escape codes are switches, turning on a given behaviour for all the
# text following it. This escape code resets all previous things turned on,
# making the text normal again
ANSI_RESET = '\33[0m'

# The 8 basic colors for the most basic escape code
BASIC_COLORS = [
	'black',
	'red',
	'green',
	'yellow',
	'blue',
	'magenta',
	'cyan',
	'white'
]

# Their RGB equivalent. A bit of cheating is involved here. As I said, the 8
# basic colors (not mixed with bold font) are quite dark. That's because their
# equivalent RGB colors are made of coordinates set at value 128, not 255.
# Here I use 255 for convenience with the rest of the program.
BASIC_RGB = [
	(0, 0, 0),
	(205, 0, 0),
	(0, 205, 0),
	(205, 205, 0),
	(0, 0, 238),
	(205, 0, 205),
	(0, 205, 205),
	(229, 229, 229)
]

# Python's re doesn't support repeated capture :s
RGB_REGEX = 'rgb\(([0-9]{1,3}),([0-9]{1,3}),([0-9]{1,3})\)'
HEXA_REGEX = '#([a-fA-F0-9]{2})([a-fA-F0-9]{2})([a-fA-F0-9]{2})'


# The following constants deals with the nature of the x-term 256 colors
# palette:
# http://www.calmar.ws/vim/256-xterm-24bit-rgb-color-chart.html
# The x-term palette is structured as follows:
#  - Codes in [0;7] are the 8 primary colors (the same supported by the most
#    basic escape code)
#  - Codes in [9;15] are also the primary colors, but the bright version (the
#    same supported by the most basic escape code when combined with bold)
#  - Codes in [16;231] are all the various colors constructed as follows: (R,
#    G, B) coordinates are incremented by regular steps. The first step is a
#    95 increment. the 4 following steps are 40 increments. (95+4*40 = 255).
#    So, for example, 16 = (0, 0, 0), 17 = (0, 0, 95), 18 = (0, 0, 135), ...,
#    21 = (0, 0, 255), 22 = (0, 95, 0), 23 = (0, 95, 95), and so on all the
#    way up to 231 = (255, 255, 255)
#  - Codes in [232;255] are gray levels (R = G = B). It starts at (8, 8, 8)
#    and increments by steps of 10, arriving at (238, 238, 238) after 23
#    steps.

# The steps
XTERM_JUMPS = [95, 40, 40, 40, 40]

# The number of code values between an increments for each of the coordinates.
# For example, 36 code values separate (0, 0, 0) from (95, 0, 0), 6 code
# values separate (0, 0, 0) from (0, 95, 0) and, of course, 1 code value
# separate (0, 0, 0) from (0, 0, 95).
# To convert an x-term code value to RGB (after having removed the
# 16-offset), you first divide by 36. The quotient gives you the number of
# steps to apply to R. You then divide the remainder by 6, which gives you the
# number of steps to apply to G. The final remainder is the number of steps to
# apply yo B.
XTERM_COEFF = [36, 6, 1]

XTERM_GRAY_LEVELS_START = 8
XTERM_GRAY_LEVELS_OFFSET = 232
XTERM_COLORS_OFFSET = 16

DEFAULT = 'default'


# The goal of this module is to allow the user to chose the palette he wants
# to use, without requiring consistency between the palette and the way the
# color is specified. For example, the user might use the x-term 256 palette
# and give a color in RGB. The module automatically makes the conversion from
# RGB to the corresponding value between 0 and 255.
# This function, given a color and a palette, returns the values to be
# inserted into the escape code corresponding to the given palette.
# It applies the following strategy: if the color is given in a format
# consistant with the palette, then it directly extracts the values from the
# color and return them. Otherwise, it converts the color to RGB and then
# converts the RGB to the format suited for the palette (or just stops at RGB
# if RGB is the format suited).
def get_color_values(color, palette):
	"""Returns a tuple of values to be inserted into the escape code
	corresponding to `palette`.

		color: a string describing a color. Accepted formats are:
		 - one of the string in `BASIC_COLORS`
		 - the decimal representation of a number between 0 and 255 (for 256
           colors palettes)
		 - a string in the form "rgb(R,G,B)" where R, G and B are integers
           between 0 and 255
		 - a string in the form "#aabbcc" where aa, bb, and cc are the
           hexadecimal representation of integers between 0 and 255

		palette: one of the keys of `PALETTES`
	"""
	rgb = None
	if color in BASIC_COLORS:
		index = BASIC_COLORS.index(color)
		if palette == '8':
			return (index,)
		else:
			rgb = tuple(255 if c > 0 else 0 for c in BASIC_RGB[index])
	elif is_in_palette(color):
		if palette == 'xterm-256':
			return (int(color),)
		else:
			rgb = xterm_palette_to_rgb(color)
	else:
		rgb_match = re.match(RGB_REGEX, color)
		if rgb_match is not None:
			rgb = rgb_match.groups()
			rgb = tuple(int(c) for c in rgb)
		else:
			hexa_match = re.match(HEXA_REGEX, color)
			if hexa_match is not None:
				rgb = hexa_match.groups()
				rgb = tuple(int(c, 16) for c in rgb)
		if (rgb_match is not None or hexa_match is not None) \
		and palette == 'rgb':
			return rgb
	if rgb is not None:
		if palette == '8':
			return (rgb_to_basic(rgb),)
		elif palette == 'xterm-256':
			return (rgb_to_xterm_palette(rgb),)
		elif palette == 'rgb':
			return rgb


def is_in_palette(color):
	try:
		value = int(color)
	except ValueError:
		return False
	return 0 <= value <= 255


def rgb_to_basic(rgb):
	bits = tuple(1 if c >= 128 else 0 for c in rgb)
	for i, color in enumerate(BASIC_RGB):
		if bits == tuple(1 if c > 0 else 0 for c in color):
			return i


def xterm_palette_to_rgb(color):
	color = int(color)
	# Basic colors
	if color <= 7:
		return BASIC_RGB[color]
	if 8 <= color <= 15:
		return tuple(255 if c > 0 else 0 for c in BASIC_RGB[color - 8])
	# Gray levels
	if XTERM_GRAY_LEVELS_OFFSET <= color <= 255:
		return (8 + 10*(color - XTERM_GRAY_LEVELS_OFFSET),) * 3
	# Other colors
	value = color - XTERM_COLORS_OFFSET
	mod = value
	rgb = []
	for c in XTERM_COEFF:
		val, mod = divmod(mod, c)
		rgb.append(sum(XTERM_JUMPS[:val]))
	return tuple(rgb)


def rgb_to_xterm_palette(rgb):
	result = None
	r, g, b = rgb
	# Gray levels
	if r == g == b:
		if r < 4:
			result = 0
		elif r > 246:
			result = 15
		elif 238 <= r <= 246:
			result = 255
		else:
			value, mod = divmod(r - XTERM_GRAY_LEVELS_START, 10)
			if mod > 4:
				value += 1
			result = XTERM_GRAY_LEVELS_OFFSET + value
	# Other colors
	total = 0
	for index, c in enumerate(rgb):
		prev_tot, tot = 0, 0
		i = 0
		while tot < c and i < len(XTERM_JUMPS):
			prev_tot = tot
			tot += XTERM_JUMPS[i]
			i += 1
		val = i if abs(tot - c) <= abs(prev_tot - c) else i-1
		total += val * XTERM_COEFF[index]
	result = XTERM_COLORS_OFFSET + total
	return result


class ColoredStr(str):

	def __new__(cls, string, color, palette='xterm-256'):
		if color == DEFAULT:
			return string
		values = get_color_values(color.lower(), palette)
		ansi_seq = get_escape(color, palette)
		literal = ansi_seq + string + ANSI_RESET
		the_string = super().__new__(cls, literal)
		the_string.length = len(string)
		the_string.true_length = len(the_string)
		the_string.lenesc = len(ansi_seq) + len(ANSI_RESET)
		the_string.original = string
		return the_string

	def __len__(self):
		return self.length

	def __iter__(self):
		return self

	def __next__(self):
		yield next(self.original)


def cstr(string, color, palette='xterm-256', no_color=False):
	if no_color:
		return string
	else:
		return ColoredStr(string, color, palette)


def get_escape(color, palette='xterm-256'):
	if color == DEFAULT:
		return None
	values = get_color_values(color.lower(), palette)
	ansi_seq = ANSI_TEMPLATES[palette].format(*values)
	return ansi_seq


if __name__ == '__main__':
	print(ColoredStr('hello, world', 'blue', 'xterm-256'))



import shutil
import os.path as op

from setuptools import setup
from setuptools.command.install import install


setup(
	name='todocli',
	version='3.4.2',
	packages=['todo', 'todo.bash_completion'],
	entry_points={
		'console_scripts': [
			'todo = todo.todo:main'
		]
	},
	include_package_data=True,
    install_requires=[
        'setuptools==60.9.3',
    ],
    extras_require={
        'tests': {
            'pyfakefs==4.5.5',
        },
    },
	author='foobuzz',
	author_email='foobuzz@fastmail.com',
	description='A command line todo list manager',
	keywords='command line todo list',
	url='https://github.com/foobuzz/todo',
	license='MIT'
)
